/* src/s_operserv.c
 *   Contains the code for oper services.
 *
 * Copyright (C) 2004-2005 Lee Hardy <lee -at- leeh.co.uk>
 * Copyright (C) 2004-2005 ircd-ratbox development team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 3.The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: s_operserv.c,v 1.15 2005/04/28 18:57:49 leeh Exp $
 */
#include "stdinc.h"

#ifdef ENABLE_OPERSERV
#include "client.h"
#include "service.h"
#include "channel.h"
#include "io.h"
#include "rserv.h"
#include "c_init.h"
#include "conf.h"
#include "hook.h"
#include "ucommand.h"
#include "modebuild.h"
#include "log.h"

static struct client *operserv_p;

static int o_oper_takeover(struct client *, struct lconn *, const char **, int);
static int o_oper_osjoin(struct client *, struct lconn *, const char **, int);
static int o_oper_ospart(struct client *, struct lconn *, const char **, int);
static int o_oper_omode(struct client *, struct lconn *, const char **, int);

static int h_operserv_sjoin_lowerts(void *chptr, void *unused);

static struct service_command operserv_command[] =
{
	{ "OSJOIN",	&o_oper_osjoin,		1, NULL, 1, 0L, 0, 0, CONF_OPER_OPERSERV, 0 },
	{ "OSPART",	&o_oper_ospart,		1, NULL, 1, 0L, 0, 0, CONF_OPER_OPERSERV, 0 },
	{ "TAKEOVER",	&o_oper_takeover,	1, NULL, 1, 0L, 0, 0, CONF_OPER_OPERSERV, 0 },
	{ "OMODE",	&o_oper_omode,		2, NULL, 1, 0L, 0, 0, CONF_OPER_OPERSERV, 0 }
};

static struct ucommand_handler operserv_ucommand[] =
{
	{ "osjoin",	o_oper_osjoin,	CONF_OPER_OPERSERV, 1, 1, NULL },
	{ "ospart",	o_oper_ospart,	CONF_OPER_OPERSERV, 1, 1, NULL },
	{ "takeover",	o_oper_takeover,CONF_OPER_OPERSERV, 1, 1, NULL },
	{ "omode",	o_oper_omode,	CONF_OPER_OPERSERV, 2, 1, NULL },
	{ "\0", NULL, 0, 0, 0, NULL }
};

static struct service_handler operserv_service = {
	"OPERSERV", "OPERSERV", "operserv", "services.int", "Oper Services",
	60, 80, operserv_command, sizeof(operserv_command), operserv_ucommand, NULL
};

static int operserv_db_callback(void *db, int, char **, char **);

void
init_s_operserv(void)
{
	operserv_p = add_service(&operserv_service);

	loc_sqlite_exec(operserv_db_callback, "SELECT * FROM operserv");

	hook_add(h_operserv_sjoin_lowerts, HOOK_SJOIN_LOWERTS);
}

static int
operserv_db_callback(void *db, int argc, char **argv, char **colnames)
{
	join_service(operserv_p, argv[0], atol(argv[1]), NULL);
	return 0;
}

static int
h_operserv_sjoin_lowerts(void *v_chptr, void *unused)
{
	struct channel *chptr = v_chptr;

	if (dlink_find(operserv_p, &chptr->services) == NULL)
		return 0;

	/* Save the new TS for later -- jilles */
	loc_sqlite_exec(NULL, "UPDATE operserv SET tsinfo = %lu "
			"WHERE chname = %Q",
			chptr->tsinfo, chptr->name);
	return 0;
}

/* preconditions: TS >= 2 and there is at least one user in the channel */
static void
otakeover(struct channel *chptr, int invite)
{
	part_service(operserv_p, chptr->name);

	remove_our_modes(chptr);

	if(invite)
		chptr->mode.mode = MODE_TOPIC|MODE_NOEXTERNAL|MODE_INVITEONLY;
	else
		chptr->mode.mode = MODE_TOPIC|MODE_NOEXTERNAL;

	chptr->tsinfo--;

	join_service(operserv_p, chptr->name, chptr->tsinfo, NULL);

	/* need to reop some services */
	if(dlink_list_length(&chptr->services) > 1)
	{
		struct client *target_p;
		dlink_node *ptr;

		modebuild_start(operserv_p, chptr);

		DLINK_FOREACH(ptr, chptr->services.head)
		{
			target_p = ptr->data;

			if(target_p != operserv_p)
				modebuild_add(DIR_ADD, "o", target_p->name);
		}

		modebuild_finish();
	}
}

static void
otakeover_full(struct channel *chptr)
{
	dlink_node *ptr, *next_ptr;

	modebuild_start(operserv_p, chptr);

	DLINK_FOREACH_SAFE(ptr, next_ptr, chptr->bans.head)
	{
		modebuild_add(DIR_DEL, "b", ptr->data);
		my_free(ptr->data);
		dlink_destroy(ptr, &chptr->bans);
	}

	DLINK_FOREACH_SAFE(ptr, next_ptr, chptr->excepts.head)
	{
		modebuild_add(DIR_DEL, "e", ptr->data);
		my_free(ptr->data);
		dlink_destroy(ptr, &chptr->excepts);
	}

	DLINK_FOREACH_SAFE(ptr, next_ptr, chptr->invites.head)
	{
		modebuild_add(DIR_DEL, "I", ptr->data);
		my_free(ptr->data);
		dlink_destroy(ptr, &chptr->invites);
	}

	modebuild_finish();
}

static void
otakeover_clear(struct channel *chptr, int remove_opers)
{
	struct chmember *msptr;
	dlink_node *ptr, *next_ptr;

	kickbuild_start();

	DLINK_FOREACH_SAFE(ptr, next_ptr, chptr->users.head)
	{
		msptr = ptr->data;

		if(msptr->client_p->user->oper ||
		   (!remove_opers && is_oper(msptr->client_p)))
			continue;

		kickbuild_add(msptr->client_p->name, "Takeover Requested");
		del_chmember(msptr);
	}

	kickbuild_finish(operserv_p, chptr);
}

static int
o_oper_takeover(struct client *client_p, struct lconn *conn_p, const char *parv[], int parc)
{
	struct channel *chptr;

	if((chptr = find_channel(parv[0])) == NULL)
	{
		service_send(operserv_p, client_p, conn_p,
				"Channel %s does not exist", parv[0]);
		return 0;
	}

	if(chptr->tsinfo < 2)
	{
		service_send(operserv_p, client_p, conn_p,
				"Channel %s TS too low for takeover", parv[0]);
		return 0;
	}

	if(dlink_list_length(&chptr->users) == 0)
	{
		/* Taking over a channel without users would lead to segfaults
		 * and is pointless anyway -- jilles */
		service_send(operserv_p, client_p, conn_p,
				"Channel %s has no users", parv[0]);
		return 0;
	}

	if(parc > 1 && !EmptyString(parv[1]))
	{
		if(!irccmp(parv[1], "-clearall"))
		{
			otakeover(chptr, 1);
			otakeover_full(chptr);
			otakeover_clear(chptr, 1);
		}
		else if(!irccmp(parv[1], "-clear"))
		{
			otakeover(chptr, 1);
			otakeover_full(chptr);
			otakeover_clear(chptr, 0);
		}
		else if(!irccmp(parv[1], "-full"))
		{
			otakeover(chptr, 0);
			otakeover_full(chptr);
		}
	}
	else
		otakeover(chptr, 0);

	slog(operserv_p, 1, "%s - TAKEOVER %s",
		OPER_NAME(client_p, conn_p), parv[0]);

	service_send(operserv_p, client_p, conn_p,
			"Channel %s has been taken over", chptr->name);
	return 0;
}

static int
o_oper_osjoin(struct client *client_p, struct lconn *conn_p, const char *parv[], int parc)
{
	struct channel *chptr;
	time_t tsinfo;

	if((chptr = find_channel(parv[0])) &&
	   dlink_find(operserv_p, &chptr->services))
	{
		service_send(operserv_p, client_p, conn_p,
				"%s already in %s", operserv_p->name, parv[0]);
		return 0;
	}

	slog(operserv_p, 1, "%s - OSJOIN %s",
		OPER_NAME(client_p, conn_p), parv[0]);

	tsinfo = chptr != NULL ? chptr->tsinfo : CURRENT_TIME;

	loc_sqlite_exec(NULL, "INSERT INTO operserv VALUES(%Q, %lu, %Q)",
			parv[0], tsinfo, OPER_NAME(client_p, conn_p));

	join_service(operserv_p, parv[0], tsinfo, NULL);

	service_send(operserv_p, client_p, conn_p,
			"%s joined to %s", operserv_p->name, parv[0]);
	return 0;
}

static int
o_oper_ospart(struct client *client_p, struct lconn *conn_p, const char *parv[], int parc)
{
	if(part_service(operserv_p, parv[0]))
	{
		slog(operserv_p, 1, "%s - OSPART %s", 
			OPER_NAME(client_p, conn_p), parv[0]);

		loc_sqlite_exec(NULL, "DELETE FROM operserv WHERE "
				"chname = %Q", parv[0]);
		service_send(operserv_p, client_p, conn_p,
				"%s removed from %s",
				operserv_p->name, parv[0]);
	}
	else
		service_send(operserv_p, client_p, conn_p,
				"%s not in channel %s", 
				operserv_p->name, parv[0]);

	return 0;
}

static int
o_oper_omode(struct client *client_p, struct lconn *conn_p, const char *parv[], int parc)
{
	struct channel *chptr;

	if((chptr = find_channel(parv[0])) == NULL)
	{
		sendto_one(conn_p, "Channel %s does not exist",	parv[0]);
		return 0;
	}

	parse_full_mode(chptr, operserv_p, parv, parc, 1);

	slog(operserv_p, 1, "%s - OMODE %s %s",
		OPER_NAME(client_p, conn_p), chptr->name,
			rebuild_params(parv, parc, 1));

	service_send(operserv_p, client_p, conn_p, "OMODE issued");
	return 0;
}

#endif
